"""Application level functions.

Functions called at the application level.

  Typical usage example:

  gene = generate_gene()
"""

import sys
import csv
import random
from itertools import combinations
from insilicodna.reference_genome import *
from insilicodna.core import chunks


def generate_gene(exon_lengths=None, n_transcripts=None):
    """Generate an instance of the Gene class.

    Generate an instance of Gene() class. If no exon lengths
    are provided, the number of exons and lengths are randomly
    sampled (uniform distribution) from a range. If number of
    transcripts are not provided, transcript number will be
    randomly sampled (uniform distrubtion) and a unique combination
    of exons will be provided.

    Args:
        exon_lengths:
          A list of ints of desired exon lengths.
        n_transcripts:
          An int of desired transcripts based on provided exons.

    Returns:
        An instance of the Gene() class

    """
    if not exon_lengths:
        n_exons = random.sample(range(1, 10), 1)[0]
        exon_lengths = random.sample(range(100, 1000), n_exons)
    # Build exons
    exons = [Exon(length=x) for x in exon_lengths]
    # Build introns
    n_introns = len(exon_lengths) - 1
    intron_lengths = random.sample(range(60, 300), n_introns)
    introns = [Intron(length=x) for x in intron_lengths]
    gene = Gene(exons=exons, introns=introns)
    gene.generate_sequence()
    # Build transcripts
    exon_combo = []
    for i in range(1, len(exon_lengths) + 1):
        combo_list = [list(x) for x in combinations(exons, i)]
        exon_combo.extend(combo_list)
    # Ensure first exon is included
    exon_combo = [x for x in exon_combo if exons[0] in x]
    if not n_transcripts:
        n_transcripts = random.sample(range(1, len(exon_lengths) + 1), 1)[0]
    transcript_combos = random.sample(exon_combo, n_transcripts)
    for exon_set in transcript_combos:
        transcript = Transcript(exon_set)
        gene.add_transcript(transcript)
    return gene

def generate_contig(
    seqid="synthetic-chromosome",
    n_genes=10,
    fasta_file=False,
    gff3_file=False,
    output_prefix="insilicodna-synthetic-data"):
    """Generate a modeled chromsome.

    Generate a modeled chromsome

    Args:
        seqid:
          A unique ID (str) to represent the contig ID.
        n_genes:
          An int of desired number of genes.

    Returns:
        An instance of the Gene() class

    """
    sequence = []
    gff3 = []
    chr_pos = 0
    for i in range(n_genes):
        intergenic = Intergenic()
        sequence.extend(intergenic.sequence)
        chr_pos += intergenic.length
        start = chr_pos + 1
        gene = generate_gene()
        sequence.extend(gene.sequence)
        chr_pos += gene.length
        attributes = f"ID={gene.synid};Name={gene.synid}"
        gff3_dict = {
            "seqid": seqid,
            "source": 'synthetic-molecular-data',
            "type": 'gene',
            "start": start,
            "end": chr_pos,
            "score": '.',
            "strand": '+',
            "phase": '.',
            "attributes": attributes
        }
        gff3.append(gff3_dict)
        # Print exon gff3
        unit_start = start
        for feature in gene.gene_structure:
            if isinstance(feature, Exon):
                featuretype = 'exon'
            elif isinstance(feature, Intron):
                featuretype = 'inton'
            else:
                pass
            unit_end = unit_start + feature.length - 1
            attributes = f"ID={feature.synid};Parent={gene.synid};Name={feature.synid}"
            gff3_dict = {
                "seqid": seqid,
                "source": 'synthetic-molecular-data',
                "type": featuretype,
                "start": unit_start,
                "end": unit_end,
                "score": '.',
                "strand": '+',
                "phase": '.',
                "attributes": attributes
            }
            unit_start = unit_end + 1
            gff3.append(gff3_dict)
        intergenic = Intergenic()
        sequence.extend(intergenic.sequence)
        chr_pos += intergenic.length

    # Print .fasta file
    if fasta_file:
        filename = f"{output_prefix}.fasta"
        seq_chunks = chunks(sequence_list=sequence, chunk_length=60)
        original_stdout = sys.stdout
        with open(filename, 'w', encoding="utf-8") as fileout:
            sys.stdout = fileout
            header = "SYNCHROMOSOME-"
            print(f">{header}")
            for line in seq_chunks:
                print("".join(line))
            sys.stdout = original_stdout

    # Print .gff3 file
    if gff3_file:
        filename = f"{output_prefix}.gff3"
        with open(filename, 'w', encoding="utf-8", ) as fileout:
            writer = csv.writer(fileout, delimiter='\t')
            writer.writerow(["##gff-version 3"])
        with open(filename, 'a', newline='', encoding="utf-8") as fileout:
            fieldnames = ['seqid', 'source', 'type', 'start',
                'end', 'score', 'strand', 'phase', 'attributes']
            writer = csv.DictWriter(fileout, fieldnames=fieldnames)
            for item in gff3:
                writer.writerow(item)
    return



"""Classes which represent features in the genome.

Classes which represent features in the genome

  Typical usage example:

  exon = Exon(length=1000)
  intron = Intron()
  transcript = Transcript()
  gene = Gene()
  intergenic = Intergenic()
"""

from insilicodna import core


class Exon:
    """Class representation of an exon.

    A simplified represention of the molecular object, exon.
    Sequences are represented in upper case.

    Attributes:
        synid (str): A unique, synthetic ID.
        length (int): Number of base pairs in the sequence.
        sequence (list of str): A list of base pairs.

    """
    def __init__(self, synid=None, length=60):
        length = max(length, 60)
        if synid is None:
            synid = f"SYNEXON-{core.generate_uuid4()}"
        self.synid = synid
        self.length = length
        self.sequence = core.generate_sequence(self.length)

class Intron:
    """Class representation of an intron.

    A simplified represention of the molecular object, intron.
    Sequences are represented in lower case.

    Attributes:
        synid (str): A unique, synthetic ID.
        length (int): Number of base pairs in the sequence.
        sequence (list of str): A list of base pairs.

    """
    def __init__(self, synid=None, length=100):
        length = max(length, 60)
        if synid is None:
            synid = f"SYNINTRON-{core.generate_uuid4()}"
        self.synid = synid
        self.length = length
        self.sequence = core.generate_sequence(self.length, upper=False)

class Transcript:
    """Class representation of a transcript.

    A simplified represention of the molecular object, transcript.

    Attributes:
        synid (str): A unique, synthetic ID.
        exons (list of obj): A list of Exon objects.
        sequence (list of str): A composite of Exon sequences,
            in order of exons in list.

    """
    def __init__(self, exons, synid=None):
        if synid is None:
            synid = f"SYNTRANSCRIPT-{core.generate_uuid4()}"
        self.synid = synid
        self.exons = exons
        self.sequence = []
        for exon in self.exons:
            self.sequence.extend(exon.sequence)

class Gene:
    """Class representation of a gene.

    A simplified represention of the molecular object, gene.

    Attributes:
        synid (str): A unique, synthetic ID.
        exons (list of obj): A list of Exon objects.
        introns (list of obj): A list of Intron objects.
        transcripts (list of obj): A list of Transcript objects.
        gene_structure (list of obj): A list of Exon and
            Intron objects.
        sequence (list of str): A list of base pairs.
        length (int): Number of base pairs in the sequence.

    """
    def __init__(self, synid=None, exons=None, introns=None, transcripts=None):
        if synid is None:
            synid = f"SYNGENE-{core.generate_uuid4()}"
        self.synid = synid
        if exons is None:
            exons = []
        self.exons = exons
        if introns is None:
            introns = []
        self.introns = introns
        if transcripts is None:
            transcripts = []
        self.transcripts = transcripts
        self.gene_structure = None
        self.sequence = None
        self.length = None

    def add_exon(self, exon):
        """Adds an Exon object to the list of exons.

        Appends an Exon object to the list of exons.

        Args:
            exon:
              A single Exon object.

        """
        self.exons.append(exon)

    def add_intron(self, intron):
        """Adds an Intron object to the list of introns.

        Appends an Intron object to the list of introns.

        Args:
            intron:
              A single Intron object.

        """
        self.exons.append(intron)

    def add_transcript(self, transcript):
        """Adds an Transcript object to the list of transcripts.

        Appends an Transcript object to the list of transcripts.

        Args:
            transcript:
              A single Transcript object.

        """
        self.transcripts.append(transcript)

    def generate_sequence(self):
        """Generates a gene sequence.

        Generates a an attribute, `gene_structure` which
        maintains a list of objects to represent a gene.

        """
        exons = self.exons
        introns = self.introns
        gene_structure = [None]*(len(exons)+len(introns))
        gene_structure[::2] = exons
        gene_structure[1::2] = introns
        sequence = []
        for item in gene_structure:
            sequence.extend(item.sequence)
        self.gene_structure = gene_structure
        self.sequence = sequence
        self.length = len(sequence)

    def generate_fasta_gene(self, filename=None):
        """Generates FASTA formated gene sequence.

        Generates FASTA formated gene sequence.

        Args:
            filename:
                Name of output FASTA file

        Returns:
            If `filename` is supplied, a file is returned,
            otherwise the FASTA sequence is printed as stdout.

        """
        exons = self.exons
        introns = self.introns
        exons_introns = [None]*(len(exons)+len(introns))
        exons_introns[::2] = exons
        exons_introns[1::2] = introns
        sequence = []
        for item in exons_introns:
            sequence.extend(item.sequence)
        fasta = [{self.synid: sequence}]
        if filename is None:
            core.print_fasta(fasta)
        else:
            core.generate_fasta_file(filename, fasta)

    def generate_fasta_transcripts(self, filename=None):
        """Generates FASTA formated transcript sequence.

        Generates FASTA formated transcript sequence.

        Args:
            filename:
                Name of output FASTA file

        Returns:
            If `filename` is supplied, a file is returned,
            otherwise the FASTA sequence is printed as stdout.

        """
        fasta = []
        for transcript in self.transcripts:
            transcript_dict = {transcript.synid: transcript.sequence}
            fasta.append(transcript_dict)
        if filename is None:
            core.print_fasta(fasta)
        else:
            core.generate_fasta_file(filename, fasta)

    def generate_fasta_exons(self, filename=None):
        """Generates FASTA formated exon sequence.

        Generates FASTA formated exon sequence.

        Args:
            filename:
                Name of output FASTA file

        Returns:
            If `filename` is supplied, a file is returned,
            otherwise the FASTA sequence is printed as stdout.

        """
        fasta = []
        for exon in self.exons:
            exon_dict = {exon.synid: exon.sequence}
            fasta.append(exon_dict)
        if filename is None:
            core.print_fasta(fasta)
        else:
            core.generate_fasta_file(filename, fasta)

class Intergenic:
    """Class representation of intergenic dna.

    A simplified represention of intergenic dna.

    Attributes:
        synid (str): A unique, synthetic ID.
        length (int): Number of base pairs in the sequence.
        sequence (list of str): A list of base pairs.

    """
    def __init__(self, synid=None, length=100):
        length = max(length, 60)
        if synid is None:
            synid = f"SYNINTERGENIC-{core.generate_uuid4()}"
        self.synid = synid
        self.length = length
        self.sequence = core.generate_sequence(self.length, upper=False)



"""Core functions which are call throughout.

Core functions which are call throughout.

"""

import random
import uuid
import sys


def chunks(sequence_list, chunk_length):
    """Yields successive n-sized chunks from a list.

    Yields successive n-sized chunks from a list. If list is not
    evenly split into equal chunks, the last chunk will contain
    the uneven chunk.

    Args:
        l:
          A single list of values to create chunks from.
        n:
          An integer which specifies the desired length of each chunk.

    Yields:
        A generator with equally sized chunks derived from the original
        list of values. Returned lists will always be equal to, or less
        than the specified input, n.

    Raises:
        IOError:
          An error occured if n is greater than the length of the
          supplied list.
    """
    for i in range(0, len(sequence_list), chunk_length):
        yield sequence_list[i:i + chunk_length]

def generate_sequence(length, upper=True):
    """Generates a sequence of base pairs.

    Generates a sequence of base pairs, randomly selecting
    from ('A', 'C', 'T', 'G').

    Args:
        length:
          Desired length of sequence.

    Returns:
        A list of base pairs

    """
    base_pairs = ['A', 'C', 'T', 'G']
    if upper is not True:
        base_pairs = [a.lower() for a in base_pairs]
    return random.choices(base_pairs, k=length)

def generate_uuid4():
    """Generates a unique id.

    Generates a unique id.

    Returns:
        An uppercase UUID of numbers (0-9) and letters (A-Z).

    """
    id_ = str(uuid.uuid4()).upper()
    id_ = id_.replace('-', '')
    return id_

def print_fasta(sequence_dict_list, line_length=60):
    """Prints a sequence in FASTA format.

    Prints a sequence in FASTA format.

    Args:
        sequence_dict_list:
          A dict where the key is an ID and the value is
          a list of base pairs.
        line_length:
          Desired line length for FASTA output.

    Returns:
        A FASTA formatted sequence.

    """
    for item in sequence_dict_list:
        header = list(item.keys())[0]
        body = list(item.values())[0]
        print(f">{header}")
        chunk_list = chunks(body, line_length)
        for line in chunk_list:
            print("".join(line))

def generate_fasta_file(filename, sequence_dict_list, line_length=60):
    """Creates a FASTA file.

    Creates a FASTA file.

    Args:
        filename:
          Name of the output file.
        sequence_dict_list:
          A dict where the key is an ID and the value is
          a list of base pairs.
        line_length:
          Desired line length for FASTA output.

    Returns:
        A FASTA formatted file.

    """
    original_stdout = sys.stdout
    filename = f"{filename}.fasta"
    with open(filename, 'w', encoding="utf-8") as fileout:
        sys.stdout = fileout
        print_fasta(sequence_dict_list, line_length=line_length)
        sys.stdout = original_stdout



"""Command line interface.

Command line accessible functions.

  Typical usage example:

  $insilicodna test
"""
import click
import insilicodna.application as app


@click.command()
@click.option('--version', is_flag=True, help="Will print version.")
@click.option('--verbose', is_flag=True, help="Will print verbose messages.")
@click.option('--gene-count', is_flag=False, default=1, help="Number of genes.")
@click.option('--output-prefix', is_flag=False, help="Assign file output prefix.")
@click.option('--fasta', is_flag=True, help="Will create the file, <prefix>.fasta.")
@click.option('--gff3', is_flag=True, help="Will create the file, <prefix>.gff3.")
def cli(
    version,
    verbose,
    gene_count,
    output_prefix,
    fasta,
    gff3):

    # Custom variables
    current_version = "v0.0.3"

    # Auto generate variables
    header = f"In Silico DNA {current_version}"
    if not output_prefix:
        output_prefix = "synthetic"

    # Print version
    if version:
        click.echo(f"{header}")
        return

    # List options
    if not fasta and not gff3:
        click.echo(f"{header}")
        click.echo("")
        click.echo("Please specify at least one of the following options:")
        click.echo("  --fasta")
        click.echo("  --gff3")
        click.echo("")
        click.echo("To list all options:")
        click.echo("  insilicodna --help")
        click.echo("")
        click.echo("Common Usage:")
        click.echo("  insilicodna --fasta --output-prefix Example")
        click.echo("    Create a synthetic file called, Example.fasta")
        click.echo("  insilicodna --gff3")
        click.echo("    Create a synthetic .gff3 file.")
        click.echo("  insilicodna --fasta --gff3")
        click.echo("    Create both a synthetic .fasta and .gff3 file.")
        click.echo("  insilicodna --fasta --gene-count 50")
        click.echo("    Create both a synthetic .fasta with 50 genes")
        return

    # Print all actions
    if verbose:
        click.echo(f"{header}")
        click.echo("")
        click.echo(f"Setting number of genes: {gene_count}")
        click.echo(f"Setting output file prefix: {output_prefix}")
        if fasta:
            click.echo(f"Creating file {output_prefix}.fasta...")
        if gff3:
            click.echo(f"Creating file {output_prefix}.gff3...")

    # Create contig
    app.generate_contig(
        seqid="synthetic-chromosome",
        n_genes=gene_count,
        fasta_file=fasta,
        gff3_file=gff3,
        output_prefix=output_prefix)

    return
